@@ -15,6 +15,7 @@
 #include "logging_p.h"
 #include <config-kglobalaccel.h>
 
+#include <KDesktopFile>
 #include <KFileUtils>
 #include <KPluginMetaData>
 
@@ -79,6 +80,82 @@ static QString getConfigFile()
     return qEnvironmentVariableIsSet("KGLOBALACCEL_TEST_MODE") ? QString() : QStringLiteral("kglobalshortcutsrc");
 }
 
+void GlobalShortcutsRegistry::migrateKHotkeys()
+{
+    KConfig hotkeys(QStringLiteral("khotkeysrc"));
+
+    int dataCount = hotkeys.group(QStringLiteral("Data")).readEntry("DataCount", 0);
+
+    for (int i = 1; i <= dataCount; ++i) {
+        const QString groupName = QStringLiteral("Data_") + QString::number(i);
+
+        KConfigGroup dataGroup(&hotkeys, groupName);
+
+        if (dataGroup.readEntry("Type") != QLatin1String("SIMPLE_ACTION_DATA")) {
+            continue;
+        }
+
+        const QString name = dataGroup.readEntry("Name");
+
+        QString exec;
+        QString uuid;
+
+        int actionsCount = KConfigGroup(&hotkeys, groupName + QLatin1String("Actions")).readEntry("ActionsCount", 0);
+
+        for (int i = 0; i < actionsCount; ++i) {
+            KConfigGroup actionGroup = KConfigGroup(&hotkeys, groupName + QLatin1String("Actions") + QString::number(i));
+            const QString type = actionGroup.readEntry("Type");
+
+            if (type == QLatin1String("COMMAND_URL")) {
+                exec = actionGroup.readEntry("CommandURL");
+            } else if (type == QLatin1String("DBUS")) {
+                exec = QStringLiteral(QT_STRINGIFY(QDBUS) " %1 %2 %3")
+                           .arg(actionGroup.readEntry("RemoteApp"), actionGroup.readEntry("RemoteObj"), actionGroup.readEntry("Call"));
+
+                const QString args = actionGroup.readEntry("Arguments");
+
+                if (!args.isEmpty()) {
+                    exec += QLatin1Char(' ') + args;
+                }
+            }
+        }
+
+        if (exec.isEmpty()) {
+            continue;
+        }
+
+        int triggerCount = KConfigGroup(&hotkeys, groupName + QLatin1String("Triggers")).readEntry("TriggersCount", 0);
+
+        for (int i = 0; i < triggerCount; ++i) {
+            KConfigGroup triggerGroup = KConfigGroup(&hotkeys, groupName + QLatin1String("Triggers") + QString::number(i));
+            if (triggerGroup.readEntry("Type") != QLatin1String("SHORTCUT")) {
+                continue;
+            }
+
+            uuid = triggerGroup.readEntry("Uuid");
+        }
+
+        const QString kglobalaccelEntry = _config.group(QStringLiteral("khotkeys")).readEntry(uuid);
+
+        if (kglobalaccelEntry.isEmpty()) {
+            continue;
+        }
+
+        const QString key = kglobalaccelEntry.split(QLatin1Char(',')).first();
+
+        KDesktopFile file(QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1String("/kglobalaccel/") + uuid
+                          + QLatin1String(".desktop"));
+        file.desktopGroup().writeEntry("Type", "Application");
+        file.desktopGroup().writeEntry("Name", name);
+        file.desktopGroup().writeEntry("Exec", exec);
+        file.desktopGroup().writeEntry("X-KDE-GlobalAccel-CommandShortcut", true);
+        file.desktopGroup().writeEntry("StartupNotify", false);
+
+        _config.group(QStringLiteral("services")).group(uuid + QLatin1String(".desktop")).writeEntry("_launch", kglobalaccelEntry);
+        _config.group(QStringLiteral("khotkeys")).revertToDefault(uuid);
+    }
+}
+
 /*
  * Migrate the config for service actions to a new format that only stores the actual shortcut if not default.
  * All other information is read from the desktop file.
@@ -111,6 +188,60 @@ void GlobalShortcutsRegistry::migrateConfig()
                 newGroup.writeEntry(key, shortcut);
             }
         }
+
+        component.deleteGroup();
+    }
+
+    // Migrate dynamic shortcuts to service-based shortcuts
+    const QStringList desktopPaths =
+        QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("kglobalaccel"), QStandardPaths::LocateDirectory);
+
+    const QStringList desktopFiles = KFileUtils::findAllUniqueFiles(desktopPaths, {QStringLiteral("*.desktop")});
+
+    for (const QString &fileName : desktopFiles) {
+        KDesktopFile file(fileName);
+        const QString componentName = QFileInfo(fileName).fileName();
+
+        auto migrateTo = [this, componentName](KConfigGroup &group, const QString &actionName, const QString displayName) {
+            QString migrateFrom = group.readEntry<QString>(QStringLiteral("X-KDE-Migrate-Shortcut"), QString());
+
+            if (migrateFrom.isEmpty()) {
+                return;
+            }
+
+            const QStringList migrateFromParts = migrateFrom.split(QLatin1Char(','));
+
+            if (!_config.group(migrateFromParts[0]).hasKey(migrateFromParts[1])) {
+                // Probably already migrated
+                return;
+            }
+
+            const QStringList shortcutTriple = _config.group(migrateFromParts[0]).readEntry<QStringList>(migrateFromParts[1], QStringList());
+            const QString oldShortcut = shortcutTriple[0];
+            const QString oldDefaultShortcut = shortcutTriple[1];
+            const QString newDefaultShortcut = group.readEntry<QString>("X-KDE-Shortcuts", QString());
+
+            // Only write value if it is not the old or new default
+            if (oldShortcut != oldDefaultShortcut && oldShortcut != newDefaultShortcut) {
+                _config.group(QStringLiteral("services")).group(componentName).writeEntry(actionName, oldShortcut);
+            }
+
+            _config.group(migrateFromParts[0]).deleteEntry(migrateFromParts[1]);
+
+            if (_config.group(migrateFromParts[0]).entryMap().size() == 1) {
+                // only _k_friendly_name left, remove the group
+                _config.deleteGroup(migrateFromParts[0]);
+            }
+        };
+
+        KConfigGroup desktopGroup = file.desktopGroup();
+        migrateTo(desktopGroup, QStringLiteral("_launch"), file.readName());
+
+        const QStringList actions = file.readActions();
+        for (const QString &action : actions) {
+            KConfigGroup actionGroup = file.actionGroup(action);
+            migrateTo(actionGroup, action, file.actionGroup(action).readEntry<QString>("Name", QString()));
+        }
     }
 
     _config.sync();
@@ -121,6 +252,7 @@ GlobalShortcutsRegistry::GlobalShortcutsRegistry()
     , _manager(loadPlugin(this))
     , _config(getConfigFile(), KConfig::SimpleConfig)
 {
+    migrateKHotkeys();
     migrateConfig();
 
     if (_manager) {
@@ -249,11 +381,9 @@ GlobalShortcutsRegistry *GlobalShortcutsRegistry::self()
     return _self;
 }
 
-/**
- * When we are provided just a Shift key press, interpret it as "Shift" not as "Shift+Shift"
- */
 static void correctKeyEvent(int &keyQt)
 {
+    // When we are provided just a Shift key press, interpret it as "Shift" not as "Shift+Shift"
     switch (keyQt) {
     case (Qt::ShiftModifier | Qt::Key_Shift).toCombined():
         keyQt = Qt::Key_Shift;
@@ -268,6 +398,15 @@ static void correctKeyEvent(int &keyQt)
         keyQt = Qt::Key_Meta;
         break;
     }
+    // Known limitation:
+    //     When shortcut is Mod(s)+Alt+Print, only works when Alt is released before Mod(s),
+    //     Does not work with multikey shortcuts.
+    // When the user presses Mod(s)+Alt+Print, the SysReq event is fired only
+    // when the Alt key is released. Before we get the Mod(s)+SysReq event, we
+    // first get a Mod(s)+Alt event, breaking multikey shortcuts.
+    if ((keyQt & ~Qt::KeyboardModifierMask) == Qt::Key_SysReq) {
+        keyQt = Qt::Key_Print | (keyQt & Qt::KeyboardModifierMask) | Qt::AltModifier;
+    }
 }
 
 bool GlobalShortcutsRegistry::keyPressed(int keyQt)
